@@ -1,26 +1,34 @@
 """Interceptor a client call with prometheus"""
+import logging
 
 from timeit import default_timer
 
 import grpc
+from prometheus_client.registry import REGISTRY
 
-import py_grpc_prometheus.grpc_utils as grpc_utils
-import py_grpc_prometheus.server_metrics
-from py_grpc_prometheus.server_metrics import GRPC_SERVER_HANDLED_HISTOGRAM
-from py_grpc_prometheus.server_metrics import GRPC_SERVER_STARTED_COUNTER
-from py_grpc_prometheus.server_metrics import GRPC_SERVER_STREAM_MSG_RECEIVED
-from py_grpc_prometheus.server_metrics import GRPC_SERVER_STREAM_MSG_SENT
-# Legacy metrics
-from py_grpc_prometheus.server_metrics import LEGACY_GRPC_SERVER_HANDLED_LATENCY_SECONDS
+from py_grpc_prometheus import grpc_utils
+from py_grpc_prometheus import server_metrics
 
 
+_LOGGER = logging.getLogger(__name__)
+
 class PromServerInterceptor(grpc.ServerInterceptor):
 
-  def __init__(self, enable_handling_time_histogram=False, legacy=False):
+  def __init__(self,
+               enable_handling_time_histogram=False,
+               legacy=False,
+               skip_exceptions=False,
+               log_exceptions=True,
+               registry=REGISTRY):
     self._enable_handling_time_histogram = enable_handling_time_histogram
     self._legacy = legacy
-    self._grpc_server_handled_total_counter = \
-      py_grpc_prometheus.server_metrics.get_grpc_server_handled_counter(self._legacy)
+    self._grpc_server_handled_total_counter = server_metrics.get_grpc_server_handled_counter(
+        self._legacy,
+        registry
+    )
+    self._metrics = server_metrics.init_metrics(registry)
+    self._skip_exceptions = skip_exceptions
+    self._log_exceptions = log_exceptions
 
   def intercept_service(self, continuation, handler_call_details):
     """
@@ -37,80 +45,75 @@ class PromServerInterceptor(grpc.ServerInterceptor):
 
     def metrics_wrapper(behavior, request_streaming, response_streaming):
       def new_behavior(request_or_iterator, servicer_context):
-        start = default_timer()
-        grpc_type = grpc_utils.get_method_type(request_streaming, response_streaming)
         try:
-          if request_streaming:
-            request_or_iterator = grpc_utils.wrap_iterator_inc_counter(
-                request_or_iterator,
-                GRPC_SERVER_STREAM_MSG_RECEIVED,
-                grpc_type,
-                grpc_service_name,
-                grpc_method_name)
-            if not self._legacy:
+          start = default_timer()
+          grpc_type = grpc_utils.get_method_type(request_streaming, response_streaming)
+          try:
+            if request_streaming:
               request_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                   request_or_iterator,
-                  GRPC_SERVER_STREAM_MSG_SENT,
+                  self._metrics["grpc_server_stream_msg_received"],
                   grpc_type,
                   grpc_service_name,
                   grpc_method_name)
-          else:
-            GRPC_SERVER_STARTED_COUNTER.labels(
-                grpc_type=grpc_type,
-                grpc_service=grpc_service_name,
-                grpc_method=grpc_method_name).inc()
+            else:
+              self._metrics["grpc_server_started_counter"].labels(
+                  grpc_type=grpc_type,
+                  grpc_service=grpc_service_name,
+                  grpc_method=grpc_method_name).inc()
 
-          # Invoke the original rpc behavior.
-          response_or_iterator = behavior(request_or_iterator, servicer_context)
+            # Invoke the original rpc behavior.
+            response_or_iterator = behavior(request_or_iterator, servicer_context)
 
-          if response_streaming:
-
-            sent_metric = GRPC_SERVER_STREAM_MSG_SENT
-            if not self._legacy:
+            if response_streaming:
+              sent_metric = self._metrics["grpc_server_stream_msg_sent"]
               response_or_iterator = grpc_utils.wrap_iterator_inc_counter(
                   response_or_iterator,
-                  GRPC_SERVER_STREAM_MSG_RECEIVED,
+                  sent_metric,
                   grpc_type,
                   grpc_service_name,
                   grpc_method_name)
 
-            response_or_iterator = grpc_utils.wrap_iterator_inc_counter(
-                response_or_iterator,
-                sent_metric,
-                grpc_type,
-                grpc_service_name,
-                grpc_method_name)
-
-          else:
+            else:
+              self.increase_grpc_server_handled_total_counter(grpc_type,
+                                                              grpc_service_name,
+                                                              grpc_method_name,
+                                                              self._compute_status_code(
+                                                                  servicer_context).name)
+            return response_or_iterator
+          except grpc.RpcError as e:
             self.increase_grpc_server_handled_total_counter(grpc_type,
                                                             grpc_service_name,
                                                             grpc_method_name,
-                                                            self._compute_status_code(
-                                                                servicer_context).name)
-          return response_or_iterator
-        except grpc.RpcError as e:
-          self.increase_grpc_server_handled_total_counter(grpc_type,
-                                                          grpc_service_name,
-                                                          grpc_method_name,
-                                                          self._compute_error_code(e).name)
+                                                            self._compute_error_code(e).name)
+            raise e
+
+          finally:
+
+            if not response_streaming:
+              if self._legacy:
+                self._metrics["legacy_grpc_server_handled_latency_seconds"].labels(
+                    grpc_type=grpc_type,
+                    grpc_service=grpc_service_name,
+                    grpc_method=grpc_method_name) \
+                    .observe(max(default_timer() - start, 0))
+              elif self._enable_handling_time_histogram:
+                self._metrics["grpc_server_handled_histogram"].labels(
+                    grpc_type=grpc_type,
+                    grpc_service=grpc_service_name,
+                    grpc_method=grpc_method_name) \
+                    .observe(max(default_timer() - start, 0))
+        except Exception as e: # pylint: disable=broad-except
+          # Allow user to skip the exceptions in order to maintain
+          # the basic functionality in the server
+          # The logging function in exception can be toggled with log_exceptions
+          # in order to suppress the noise in logging
+          if self._skip_exceptions:
+            if self._log_exceptions:
+              _LOGGER.error(e)
+            return response_or_iterator
           raise e
 
-        finally:
-
-          if not response_streaming:
-            if self._legacy:
-              LEGACY_GRPC_SERVER_HANDLED_LATENCY_SECONDS.labels(
-                  grpc_type=grpc_type,
-                  grpc_service=grpc_service_name,
-                  grpc_method=grpc_method_name) \
-                  .observe(max(default_timer() - start, 0))
-            elif self._enable_handling_time_histogram:
-              GRPC_SERVER_HANDLED_HISTOGRAM.labels(
-                  grpc_type=grpc_type,
-                  grpc_service=grpc_service_name,
-                  grpc_method=grpc_method_name) \
-                  .observe(max(default_timer() - start, 0))
-
       return new_behavior
 
     optional_any = self._wrap_rpc_behavior(continuation(handler_call_details), metrics_wrapper)
